#include <stdio.h>
#include <string.h>
#include "messages.h"
#include "darray.h"
#include "stg_net.h"
#include "str_utils.h"
#include "stg.h"

const char* KEY_MODEL            = ".model";
const char* KEY_INPUTS           = ".inputs";
const char* KEY_OUTPUTS          = ".outputs";
const char* KEY_INTERNAL         = ".internal";
const char* KEY_DUMMY            = ".dummy";
const char* KEY_INITIAL_STATE    = ".initial state";
const char* KEY_GRAPH            = ".graph";
const char* KEY_MARKING          = ".marking";
const char* KEY_CAPACITY         = ".capacity";
const char* KEY_END              = ".end";

const char*  PATTERN_HIGH        = "%s";
const char*  PATTERN_LOW         = "!%s";
const char* SYMBOLS_POLARITY = "+-";

const char* PATTERN_PLACE = "%s%d";
const char* NAME_PLACE = "p";

//---------------------------------------------------------------------------
STG::STG(char* name, SignalArray* signals, TransitionArray* transitions, 
  PlaceArray* places) : Object(tag, name)
{
  this->signal_hash_table = new SignalHashTable();
  this->signals = new SignalArray();
  this->addSignals(signals);
  
  this->transition_hash_table = new TransitionHashTable();
  this->transitions = new TransitionArray();
  this->addTransitions(transitions);

  this->place_hash_table = new PlaceHashTable();
  this->places = new PlaceArray();
  this->addPlaces(places);
}

//---------------------------------------------------------------------------
STG::STG(const STG& stg) : Object(stg)
{
  signal_hash_table = new SignalHashTable();
  signals = new SignalArray(stg.signals->num);
  for (int i = 0; i < stg.signals->num; i++)
    addSignal(new Signal(*stg.signals->arr[i]));
  
  transition_hash_table = new TransitionHashTable();
  transitions = new TransitionArray(stg.transitions->num);
  for (int i = 0; i < stg.transitions->num; i++)
    addTransition(new Transition(*stg.transitions->arr[i]));
  
  place_hash_table = new PlaceHashTable();
  places = new PlaceArray(stg.places->num);
  for (int i = 0; i < stg.places->num; i++)
    addPlace(new Place(*stg.places->arr[i]));
  
  calcIncidence();
}

//---------------------------------------------------------------------------
void STG::calcIncidence()
{
  for (int i = 0; i < transitions->num; i++)
  {
    Transition* transition_cur = transitions->arr[i];
    calcIncidence(transition_cur);
  }
  for (int i = 0; i < places->num; i++)
  {
    Place* place_cur = places->arr[i];
    calcIncidence(place_cur);
  }
}

//---------------------------------------------------------------------------
void STG::calcIncidence(Transition* transition)
{
  for (int i = 0; i < transition->preset->num; i++)
  {
    Place* place_cur = getPlace(transition->preset->arr[i]->getName());
    transition->preset->arr[i] = place_cur;
  }
  for (int i = 0; i < transition->postset->num; i++)
  {
    Place* place_cur = getPlace(transition->postset->arr[i]->getName());
    transition->postset->arr[i] = place_cur;
  }
  for (int i = 0; i < transition->readset->num; i++)
  {
    Place* place_cur = getPlace(transition->readset->arr[i]->getName());
    transition->readset->arr[i] = place_cur;
  }
}
 
//---------------------------------------------------------------------------
void STG::calcIncidence(Place* place)
{
  for (int i = 0; i < place->preset->num; i++)
  {
    Transition* transition_cur = getTransition(place->preset->arr[i]->getName());
    place->preset->arr[i] = transition_cur;
  }
  for (int i = 0; i < place->postset->num; i++)
  {
    Transition* transition_cur = getTransition(place->postset->arr[i]->getName());
    place->postset->arr[i] = transition_cur;
  }
  for (int i = 0; i < place->readset->num; i++)
  {
    Transition* transition_cur = getTransition(place->readset->arr[i]->getName());
    place->readset->arr[i] = transition_cur;
  }
}

//---------------------------------------------------------------------------
STG::~STG()
{
  clean();
  
  delete signal_hash_table; signal_hash_table = NULL;
  delete signals; signals = NULL;
  
  delete transition_hash_table; transition_hash_table = NULL;
  delete transitions; transitions = NULL;

  delete place_hash_table; place_hash_table = NULL;
  delete places; places = NULL;
}

//---------------------------------------------------------------------------
void STG::clean()
{
  if (signals) signals->removeAll(true);
  if (transitions) transitions->removeAll(true);
  if (places) places->removeAll(true);
}

//---------------------------------------------------------------------------
Signal* STG::addSignal(char* name)
{
  Signal* signal_new = new Signal( namecopy(name) );
  Signal* signal = signal_hash_table->find(signal_new);
  if (signal)
  	delete signal_new;
  else
  {
  	signal = signal_new;
   	signals->add(signal);
   	signal_hash_table->add(signal);
  }
  return signal;
}

//---------------------------------------------------------------------------
bool STG::addSignal(Signal* signal)
{
  bool result = false;
  if (!signal_hash_table->find(signal))
  {
   	signals->add(signal);
   	signal_hash_table->add(signal);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void STG::addSignals(SignalArray* signals)
{
  if (signals)
  for (int i = 0; i < signals->num; i++)
  {
  	Signal* signal_cur = signals->arr[i];
  	if (!addSignal(signal_cur))
      delete signal_cur;
  }
}

//---------------------------------------------------------------------------
void STG::removeSignal(Signal* signal)
{
  signal_hash_table->remove(signal);
  signals->remove(signal, true);
}

//---------------------------------------------------------------------------
Signal* STG::getSignal(char* name)
{
  char* signal_name = namecopy(name);
  char* suffix_str = strpbrk(signal_name, "+-");
  if (suffix_str)
    suffix_str[0] = '\0';
  Signal* signal_mask = new Signal(signal_name);
  Signal* signal_cur = signal_hash_table->find(signal_mask);
  delete signal_mask;
  return signal_cur;
}

//---------------------------------------------------------------------------
Transition* STG::addTransition(char* name)
{
  Transition* transition_new = new Transition(namecopy(name), this->signals);
  Transition* transition = transition_hash_table->find(transition_new);
  if (transition)
  	delete transition_new;
  else
  {
  	transition = transition_new;
   	transitions->add(transition);
   	transition_hash_table->add(transition);
  }
  return transition;
}

//---------------------------------------------------------------------------
bool STG::addTransition(Transition* transition)
{
  bool result = false;
  if (!transition_hash_table->find(transition))
  {
   	transitions->add(transition);
   	transition_hash_table->add(transition);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void STG::addTransitions(TransitionArray* transitions)
{
  if (transitions)
  for (int i = 0; i < transitions->num; i++)
  {
  	Transition* transition_cur = transitions->arr[i];
  	if (!addTransition(transition_cur))
      delete transition_cur;
  }
}

//---------------------------------------------------------------------------
void STG::removeTransition(Transition* transition)
{
  for (int i = 0; i < places->num; i++)
  {
    Place* place_cur = places->arr[i];
    while ( isConsumingArc(place_cur, transition) )
      removeConsumingArc(place_cur, transition);
    while ( isProducingArc(transition, place_cur) )
      removeProducingArc(transition, place_cur);
    while ( isReadArc(place_cur, transition) )
      removeReadArc(place_cur, transition);
  }
  transition_hash_table->remove(transition);
  transitions->remove(transition, true);
}

//---------------------------------------------------------------------------
Transition* STG::getTransition(char* name)
{
  Transition* transition_mask = new Transition( namecopy(name), this->signals );
  Transition* transition_cur = transition_hash_table->find(transition_mask);
  delete transition_mask;
  return transition_cur;
}

//---------------------------------------------------------------------------
Place* STG::addPlace(char* name)
{
  Place* place_new = new Place(namecopy(name));
  Place* place = place_hash_table->find(place_new);
  if (place)
  	delete place_new;
  else
  {
  	place = place_new;
   	places->add(place);
   	place_hash_table->add(place);
  }
  return place;
}

//---------------------------------------------------------------------------
bool STG::addPlace(Place* place)
{
  bool result = false;
  if (!place_hash_table->find(place))
  {
   	places->add(place);
   	place_hash_table->add(place);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void STG::addPlaces(PlaceArray* places)
{
  if (places)
  for (int i = 0; i < places->num; i++)
  {
  	Place* place_cur = places->arr[i];
  	if (!addPlace(place_cur))
      delete place_cur;
  }
}

//---------------------------------------------------------------------------
void STG::removePlace(Place* place)
{
  for (int i = 0; i < transitions->num; i++)
  {
    Transition* transition_cur = transitions->arr[i];
    while ( isConsumingArc(place, transition_cur) )
      removeConsumingArc(place, transition_cur);
    while ( isProducingArc(transition_cur, place) )
      removeProducingArc(transition_cur, place);
    while ( isReadArc(place, transition_cur) )
      removeReadArc(place, transition_cur);
  }
  place_hash_table->remove(place);
  places->remove(place, true);
}

//---------------------------------------------------------------------------
Place* STG::getPlace(char* name)
{
  Place* place_mask = new Place( namecopy(name) );
  Place* place_cur = place_hash_table->find(place_mask);
  delete place_mask;
  return place_cur;
}


//---------------------------------------------------------------------------
Place* STG::getPlace(char* name_preset, char* name_postset)
{
  Place* place = NULL;
  Transition* transition_preset = getTransition(name_preset);
  Transition* transition_postset = getTransition(name_postset);
  for (int i = 0; i < places->num; i++)
  {
    Place* place_cur = places->arr[i];
    if ( isProducingArc(transition_preset, place_cur) 
      && isConsumingArc(place_cur, transition_postset) )
      place = place_cur;
  }
  return place;
}

//---------------------------------------------------------------------------
char* STG::genPlaceName()
{
  int i = 0;
  char* name_tmp = name2pattern(PATTERN_PLACE, NAME_PLACE, i);
  while ( getPlace(name_tmp) )
  {
    delete [] name_tmp;
    i++;
    name_tmp = name2pattern(PATTERN_PLACE, NAME_PLACE, i);
  }
  return name_tmp;
}

//---------------------------------------------------------------------------
void STG::addConsumingArc(Place* place, Transition* transition)
{
   place->postset->add(transition);
   transition->preset->add(place);
}

//---------------------------------------------------------------------------
void STG::removeConsumingArc(Place* place, Transition* transition)
{
  place->postset->remove(transition, false);
  transition->preset->remove(place, false);
}

//---------------------------------------------------------------------------
bool STG::isConsumingArc(Place* place, Transition* transition)
{
  return ( (place->postset->locate(transition) >= 0)
        && (transition->preset->locate(place) >= 0) );
}

//---------------------------------------------------------------------------
void STG::addProducingArc(Transition* transition, Place* place)
{
   transition->postset->add(place);
   place->preset->add(transition);
}

//---------------------------------------------------------------------------
void STG::removeProducingArc(Transition* transition, Place* place)
{
  transition->postset->remove(place, false);
  place->preset->remove(transition, false);
}

//---------------------------------------------------------------------------
bool STG::isProducingArc(Transition* transition, Place* place)
{
  return ( (transition->postset->locate(place) >= 0)
        && (place->preset->locate(transition) >= 0) );
}

//---------------------------------------------------------------------------
void STG::addReadArc(Place* place, Transition* transition)
{
   place->readset->add(transition);
   transition->readset->add(place);
}

//---------------------------------------------------------------------------
void STG::removeReadArc(Place* place, Transition* transition)
{
   place->readset->remove(transition, false);
   transition->readset->remove(place, false);
}

//---------------------------------------------------------------------------
bool STG::isReadArc(Place* place, Transition* transition)
{
  return ( (place->readset->locate(transition) >= 0)
        && (transition->readset->locate(place) >= 0) );
}

//---------------------------------------------------------------------------
void STG::addArc(Node* from, Node* to)
{
  Signal* signal_from = getSignal(from->getName());
  Signal* signal_to = getSignal(to->getName());
  if (signal_from && signal_to)
  {
    Node* middle = new Node( genPlaceName() );
    addArc(from, middle);
    addArc(middle, to);
    delete middle;
  }
  else
  if (signal_from)
  {
  	// transition -> place
    Transition* transition = addTransition(from->getName());
    Place* place = addPlace(to->getName());
    if ( isConsumingArc(place, transition) )
    {
      removeConsumingArc(place, transition);
      if ( ! isReadArc(place, transition) )
        addReadArc(place, transition);
    }
    else
      addProducingArc(transition, place);
  }
  else
  if (signal_to)
  {
  	// place -> transition
    Place* place = addPlace(from->getName());
    Transition* transition = addTransition(to->getName());
    if ( isProducingArc(transition, place) )
    {
      removeProducingArc(transition, place);
      if ( ! isReadArc(place, transition) )
        addReadArc(place, transition);
    }
    else
      addConsumingArc(place, transition);
  }
}

//---------------------------------------------------------------------------
void STG::readASTG(char* file_names)
{
  if (file_names)
  {
    int name_index = 0;
    stg_net = this;
    while (stg_file_name = names2name(file_names, name_index++))
    {
      if (strcmp(stg_file_name, NAME_STDIO) == 0)
      {
        delete [] stg_file_name;
        stg_file_name = namecopy("stdin");
        stg_in = stdin;
      }
      else
      if ((stg_in = fopen(stg_file_name, "r")) == NULL)
        Error("opening input file <%s>", stg_file_name);
      stg_parse();
      if (stg_in != stdin)
        fclose(stg_in);
      delete [] stg_file_name;
    }
  }
}

//---------------------------------------------------------------------------
void STG::writeASTG(char* file_name)
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
	  bool label_printed = false;
	
	  // .model
	  if (name)
	    fprintf(file, "%s %s\n", KEY_MODEL, name);
	
	  // .inputs
	  label_printed = false;
	  for (int i = 0; i < signals->num; i++)
	  {
	  	Signal* signal_cur = signals->arr[i];
	    if (signal_cur->getType() == Signal::TYPE_INPUT)
	    {
	      if (!label_printed)
	      {
	        fprintf(file, KEY_INPUTS);
	        label_printed = true;
	      }
	      fprintf(file, " %s", signal_cur->getName());
	    }
	  }
	  if (label_printed)
	    fprintf(file, "\n");
	
	  // .outputs
	  label_printed = false;
	  for (int i = 0; i < signals->num; i++)
	  {
	  	Signal* signal_cur = signals->arr[i];
	    if (signal_cur->getType() == Signal::TYPE_OUTPUT)
	    {
	      if (!label_printed)
	      {
	        fprintf(file, KEY_OUTPUTS);
	        label_printed = true;
	      }
	      fprintf(file, " %s", signal_cur->getName());
	    }
	  }
	  if (label_printed)
	    fprintf(file, "\n");
	
	  // .internal
	  label_printed = false;
	  for (int i = 0; i < signals->num; i++)
	  {
	  	Signal* signal_cur = signals->arr[i];
	    if (signal_cur->getType() == Signal::TYPE_INTERNAL)
	    {
	      if (!label_printed)
	      {
	        fprintf(file, KEY_INTERNAL);
	        label_printed = true;
	      }
	      fprintf(file, " %s", signal_cur->getName());
	    }
	  }
	  if (label_printed)
	    fprintf(file, "\n");
	
	  // .dummy
	  label_printed = false;
	  for (int i = 0; i < signals->num; i++)
	  {
	  	Signal* signal_cur = signals->arr[i];
	    if (signal_cur->getType() == Signal::TYPE_DUMMY)
	    {
	      if (!label_printed)
	      {
	        fprintf(file, KEY_DUMMY);
	        label_printed = true;
	      }
	      fprintf(file, " %s", signals->arr[i]->getName());
	    }
	  }
	  if (label_printed)
	    fprintf(file, "\n");
	
	  //.initial state
	  label_printed = false;
	  for (int i = 0; i < signals->num; i++)
	  {
	  	Signal* signal_cur = signals->arr[i];
	    if ( (signal_cur->getState() == Signal::STATE_LOW)
	      || (signal_cur->getState() == Signal::STATE_HIGH) )
	    {
	      if (!label_printed)
	      {
	        fprintf(file, KEY_INITIAL_STATE);
	        label_printed = true;
	      }
	      fprintf(file, " ");
	      if (signal_cur->getState() == Signal::STATE_LOW)
	        fprintf( file, PATTERN_LOW, signal_cur->getName() );
	      if (signal_cur->getState() == Signal::STATE_HIGH)
	        fprintf( file, PATTERN_HIGH, signal_cur->getName() );
	    }
	  }
	  if (label_printed)
	    fprintf(file, "\n");
	  //.graph
	  fprintf(file, "%s\n", KEY_GRAPH);
	  // producing arcs
      if (transitions)
	  for (int i = 0; i < transitions->num; i++)
	  {
	    Transition* transition_cur = transitions->arr[i];
        if (transition_cur->postset->num > 0)
	    {
	      fprintf(file, "%s", transition_cur->getName());
          for (int j = 0; j < transition_cur->postset->num; j++)
            fprintf( file, " %s", transition_cur->postset->arr[j]->getName() );
          fprintf(file, "\n");
	    }
	  }
	  // consuming arcs
	  for (int i = 0; i < places->num; i++)
	  {
	    Place* place_cur = places->arr[i];
	    if (place_cur->postset)
	      if (place_cur->postset->num > 0)
	      {
	        fprintf(file, "%s", place_cur->getName());
            for (int j = 0; j < place_cur->postset->num; j++)
              fprintf(file, " %s", place_cur->postset->arr[j]->getName());
            fprintf(file, "\n");
	      }
	  }
	  // read-arcs
	  for (int i = 0; i < places->num; i++)
	  {
	    Place* place_cur = places->arr[i];
        if (place_cur->readset->num > 0)
	    {
	      for (int j = 0; j < place_cur->readset->num; j++)
	      {
	        Transition* transition_cur = place_cur->readset->arr[j];
	        fprintf(file, "%s %s\n", place_cur->getName(), transition_cur->getName());
            fprintf(file, "%s %s\n", transition_cur->getName(), place_cur->getName());
	      }
	    }
	  }
	  //.marking
	  fprintf(file, "%s {", KEY_MARKING);
	  label_printed = false;
	  for (int i = 0; i < places->num; i++)
	  {
	    Place* place_cur = places->arr[i];
	    if (place_cur->getMarking() > 0)
	    {
	      if (label_printed)
	        fprintf(file, " ");
        fprintf(file, "%s", place_cur->getName());
	      label_printed = true;
	    }
	  }
	  fprintf(file, "}\n");
	  //.end
	  fprintf(file, "%s\n", KEY_END);
    if (file != stdout)
      fclose(file);
  }
}
