#include "sv_circuit.h"
#include "sv_object.h"
#include "sv_module.h"
#include "sv_lib.h"
#include "sverilog.h"
#include "messages.h"

//---------------------------------------------------------------------------
Circuit::Circuit(int tag, char* name, ModuleArray* modules, Lib* lib) : VerilogObject(tag, name)
{
  this->lib = lib;
  module_hash_table = new ModuleHashTable();
  this->modules = NULL;
  this->addModules(modules);
}

//---------------------------------------------------------------------------
Circuit::Circuit(const Circuit &circuit) : VerilogObject(circuit)
{
  this->lib = circuit.lib;
  this->modules = NULL;
  this->addModules(circuit.modules);
}

//---------------------------------------------------------------------------
Circuit::~Circuit()
{
  clean();
  delete this->module_hash_table; this->module_hash_table = NULL;
  delete this->modules; this->modules = NULL;
}

//---------------------------------------------------------------------------
void Circuit::clean()
{
  // modules are actualy stored in the library, 
  // so there is no need to clean the modules array
}

//---------------------------------------------------------------------------
void Circuit::readModules(char* file_names)
{
  if (file_names)
  {
    int name_index = 0;
    sverilog_modules = new ModuleArray(1);
    while (sverilog_filename = names2name(file_names, name_index++))
    {
      if (strcmp(sverilog_filename, NAME_STDIO) == 0)
      {
        delete [] sverilog_filename;
        sverilog_filename = namecopy("stdin");
        sverilog_in = stdin;
      }
      else
      if ((sverilog_in = fopen(sverilog_filename, "r")) == NULL)
        Error("opening input file <%s>", sverilog_filename);
      sverilog_parse();
      if (sverilog_in != stdin)
        fclose(sverilog_in);
      delete [] sverilog_filename;
    }
    this->addModules(sverilog_modules);
    delete sverilog_modules; sverilog_modules = NULL;
  }
  this->lib->calcConnections(this->modules);
  this->lib->sortConnections(this->modules);
}


//---------------------------------------------------------------------------
void Circuit::writeModules(char* file_name)
{
  this->lib->writeModules(file_name, this->modules);
}

//---------------------------------------------------------------------------
void Circuit::addModule(Module* module, int position)
{
  if (this->module_hash_table->find(module) == NULL)
  {
    if (!this->modules)
      this->modules = new ModuleArray(1);
    module->lib = this->lib;
    if (position < 0)
      this->modules->add(module);
    else
      this->modules->addAt(position, module);
    this->module_hash_table->add(module);
    this->lib->addModule(module);
  }
  else
  {
    Warning("module <%s> repeats in the circuit", module->getName());
    delete module;
  }
}

//---------------------------------------------------------------------------
void Circuit::addModules(ModuleArray* modules)
{
  if (modules)
  for (int module_index = 0; module_index < modules->num; module_index++)
  {
    Module* module_cur = modules->arr[module_index];
    this->addModule(module_cur);
  }
}

//---------------------------------------------------------------------------
void Circuit::removeModule(Module* module)
{
  this->module_hash_table->remove(module);
  this->modules->remove(module, false);
}

//---------------------------------------------------------------------------
Module* Circuit::getModule(char* module_name)
{
  Module* module_mask = new Module(TAG_UNDEFINED, namecopy(module_name));
  Module* module_cur = module_hash_table->find(module_mask);
  delete module_mask;
  return module_cur;
}

//---------------------------------------------------------------------------
Module* Circuit::setModuleTop(char* module_name)
{
  // try to find topmost module by name
  this->module_top = this->getModule(module_name);
  if (module_top == NULL)
    Error("unable to find the topmost module <%s>", module_name);
  return this->module_top;
}

//---------------------------------------------------------------------------
Module* Circuit::getModuleTop()
{
  return this->module_top;
}

//---------------------------------------------------------------------------
Module* Circuit::calcModuleTop()
{
  module_top = NULL;
  ModuleArray* modules_internal = new ModuleArray(1);
  if (this->modules)
  for (int module_index = 0; module_index < this->modules->num; module_index++)
  {
    Module* module_cur = this->modules->arr[module_index];
    if (module_cur->instances)
    for (int instance_index = 0; instance_index < module_cur->instances->num; instance_index++)
    {
      Instance* instance_cur = module_cur->instances->arr[instance_index];
      Module* module_internal = this->getModule(instance_cur->getModuleName());
      if (module_internal)
        modules_internal->add(module_internal);
    }
  }
  // if there is at least 1 not internal module
  if (this->modules)
  for (int module_index = 0; module_index < this->modules->num; module_index++)
  {
    Module* module_cur = this->modules->arr[module_index];
    if (modules_internal->locate(module_cur) == -1)
      if (module_top)
      {
        module_top = NULL;
        break;
      }
      else
        module_top = module_cur;
  }
  delete modules_internal; modules_internal = NULL;
  if (module_top)
    Message("the topmost module <%s> detected", module_top->getName());
  else
    Error("unable to detect the topmost module");
  return this->module_top;
}

//---------------------------------------------------------------------------
void Circuit::calcHierarchy()
{
  this->lib->calcHierarchy(this->modules);
}

//---------------------------------------------------------------------------
int Circuit::getMaxDepth()
{
  return this->lib->getMaxDepth(this->modules);
}

//---------------------------------------------------------------------------
bool Circuit::isHierarchical()
{
  return (this->getMaxDepth() == 0);
}
