#include <stdio.h>
#include "sv_lib.h"
#include "sverilog.h"
#include "messages.h"

//---------------------------------------------------------------------------
Lib::Lib(ModuleArray* modules)
{
  // modules
  this->module_hash_table = new ModuleHashTable();
  this->modules = modules;
  if (this->modules)
  for (int module_index = 0; module_index < this->modules->num; module_index++)
    module_hash_table->add(this->modules->arr[module_index]);
}

//---------------------------------------------------------------------------
Lib::~Lib()
{
  clean();
  delete module_hash_table;
  delete modules;
}

//---------------------------------------------------------------------------
void Lib::clean()
{
  if (this->modules) this->modules->removeAll(true);
}

//---------------------------------------------------------------------------
void Lib::readModules(char* file_names)
{
  if (file_names)
  {
    int name_index = 0;
    while (sverilog_filename = names2name(file_names, name_index++))
    {
      if (strcmp(sverilog_filename, NAME_STDIO) == 0)
      {
        delete [] sverilog_filename;
        sverilog_filename = namecopy("stdin");
        sverilog_in = stdin;
      }
      else
      if ((sverilog_in = fopen(sverilog_filename, "r")) == NULL)
        Error("opening library file <%s>", sverilog_filename);

      if (sverilog_in)
      {
        Message("reading file <%s>", sverilog_filename);
        sverilog_modules = new ModuleArray(1);
        sverilog_parse();
        if (sverilog_in != stdin)
          fclose(sverilog_in);
        this->addModules(sverilog_modules);
        delete sverilog_modules;
      }
      delete [] sverilog_filename;
    }
    this->calcConnections(this->modules);
    this->sortConnections(this->modules);
  }
}

//---------------------------------------------------------------------------
void Lib::writeModule(FILE* file, Module* module_cur, int tag_ignore)
{
  if (!module_cur->testTag(tag_ignore))
  if (module_cur->instances)
  {
        // module
        fprintf(file, "%s %s", KEY_MODULE, module_cur->getName());
        int line_len = 6 + 2 + 2;
        if (module_cur->getName())
          line_len += strlen(module_cur->getName());

        // port list
        bool port_first = true;
        if (module_cur->ports)
        for (int j=0; j < module_cur->ports->num; j++)
        {
          Port* port_cur = module_cur->ports->arr[j];
          if (!port_cur->testTag(TAG_OBJ_HIDDEN))
          {
            if (port_cur->testTag(TAG_PORT_INPUT) ||
              port_cur->testTag(TAG_PORT_OUTPUT))
            {
              if (port_first)
                fprintf(file, " (");
              else
                fprintf(file, ", ");

              char* port_name = NULL;
              if (port_cur->isEscape())
                port_name = name2pattern("%s ", port_cur->getName());
              else
                port_name = namecopy(port_cur->getName());

              if (line_len + strlen(port_name) > LINE_LEN_MAX)
              {
                line_len = 2;
                fprintf(file, "\n  %s", port_name);
              }
              else
                fprintf(file, "%s", port_name);
              line_len += 2 + strlen(port_name);
              delete [] port_name;
              port_first = false;
            }
          }
        }
        fprintf(file, ");\n");

        // port declaration
        bool port_bl = false;
        if (module_cur->ports)
        for (int j=0; j < module_cur->ports->num; j++)
        {
          Port* port_cur = module_cur->ports->arr[j];
          if (!port_cur->testTag(TAG_OBJ_HIDDEN))
          {
            port_bl = true;
            switch (port_cur->getTag() & TAG_PORT_)
            {
              case TAG_PORT_INPUT:
                fprintf(file, "  %s", KEY_INPUT);
                break;
              case TAG_PORT_OUTPUT:
                fprintf(file, "  %s", KEY_OUTPUT);
                break;
              case TAG_PORT_WIRE:
                fprintf(file, "  %s", KEY_WIRE);
                break;
              default:
                fprintf(file, "  TAG_%X", port_cur->getTag());
            }
            char* port_full_name = port_cur->genFullName();
            fprintf(file, " %s;\n", port_full_name);
            delete [] port_full_name;
          }
        }
        if (port_bl)
          fprintf(file, "\n");

        // parameters
        bool parameter_bl = false;
        if (module_cur->parameters)
        for (int j=0; j < module_cur->parameters->num; j++)
        {
          Parameter* parameter_cur = module_cur->parameters->arr[j];
          parameter_bl = true;
          if (parameter_cur->testTag(TAG_OBJ_HIDDEN))
            fprintf(file, "//");
          fprintf(file, "  %s %s = %g;\n", KEY_PARAMETER, parameter_cur->getName(), parameter_cur->getValue());
        }
        if (parameter_bl)
          fprintf(file, "\n");

        // instances
        if (module_cur->instances)
        for (int j=0; j < module_cur->instances->num; j++)
        {
          Instance* instance_cur = module_cur->instances->arr[j];
          char* instance_name = instance_cur->genFullName();
          fprintf(file, "  %s", instance_cur->getModuleName());
          if (instance_name)
            fprintf(file, " %s", instance_name);
          int line_len = 2 + 1 + 2;
          if (instance_cur->getModuleName())
            line_len += strlen(instance_cur->getModuleName());
          if (instance_name)
            line_len += strlen(instance_name);
          delete [] instance_name;

          // connection list
          bool connection_first = true;
          if (instance_cur->connections)
          for (int k=0; k < instance_cur->connections->num; k++)
          {
            Connection* connection_cur = instance_cur->connections->arr[k];
            if (connection_first)
              fprintf(file, " (");
            else
              fprintf(file, ", ");
            char* connection_full_name = connection_cur->genFullName();
            if (line_len + strlen(connection_full_name) > LINE_LEN_MAX)
            {
              line_len = 4;
              fprintf(file, "\n    %s", connection_full_name);
            }
            else
              fprintf(file, "%s", connection_full_name);
            line_len += 2 + strlen(connection_full_name);
            delete [] connection_full_name;
            connection_first = false;
          }
          fprintf(file, ");\n");
        }
        // endmodule
        fprintf(file, "%s\n\n", KEY_ENDMODULE);
  }
}

//---------------------------------------------------------------------------
void Lib::writeModules(char* file_name, ModuleArray* modules)
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    if (modules)
    for (int module_index = 0; module_index < modules->num; module_index++)
    {
      Module* module_cur = modules->arr[module_index];
      writeModule(file, module_cur);
    }
    if (file != stdout)
      fclose(file);
  }
}

//---------------------------------------------------------------------------
bool Lib::addModule(Module* module, int position)
{
  bool result = false;
  if (!this->module_hash_table->find(module))
  {
    if (!this->modules)
      this->modules = new ModuleArray(1);
    module->lib = this;
    // add module to the library
    this->modules->add(module);
    this->module_hash_table->add(module);
    result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void Lib::addModules(ModuleArray* modules)
{
  if (modules)
  for (int module_index = 0; module_index < modules->num; module_index++)
  {
    Module* module_cur = modules->arr[module_index];
    if (!this->addModule(module_cur))
      delete module_cur;
  }
}

//---------------------------------------------------------------------------
Module* Lib::getModule(char* module_name)
{
  Module* module_mask = new Module(TAG_UNDEFINED, namecopy(module_name));
  Module* module_cur = module_hash_table->find(module_mask);
  delete module_mask;
  return module_cur;
}

//---------------------------------------------------------------------------
void Lib::calcConnections(ModuleArray* modules)
{
  if (modules)
  for (int module_index = 0; module_index < modules->num; module_index++)
  {
    Module* module_cur = modules->arr[module_index];
    if (module_cur->testTag(TAG_MODULE_COMPLEX))
      module_cur->calcConnections();
  }
}

//---------------------------------------------------------------------------
void Lib::sortConnections(ModuleArray* modules)
{
  if (modules)
  for (int module_index = 0; module_index < modules->num; module_index++)
  {
    Module* module_cur = modules->arr[module_index];
    if (module_cur->testTag(TAG_MODULE_COMPLEX))
      module_cur->sortConnections();
  }
}


//---------------------------------------------------------------------------
void Lib::findHierarchy(ModuleArray* modules, Module* module, int depth)
{
  if (!module->testTag(TAG_OBJ_PROCESSED))
  if (depth > module->getDepth())
  {
    module->setDepth(depth);
    module->orTag(TAG_OBJ_PROCESSED);
    if (modules)
    for (int module_index = 0; module_index < modules->num; module_index++)
    {
      Module* module_cur = modules->arr[module_index];
      if (module_cur->users)
      for (int user_index = 0; user_index < module_cur->users->num; user_index++)
      {
      	Instance* user_cur = module_cur->users->arr[user_index];
      	if (user_cur->getContainer() == module)
          this->findHierarchy(modules, module_cur, depth+1);
      }
    }
    module->andTag(~TAG_OBJ_PROCESSED);
  }
}


//---------------------------------------------------------------------------
void Lib::calcHierarchy(ModuleArray* modules)
{
  if (modules == NULL)
    modules = this->modules;
  if (modules)
  {
    // clear depth and processed flag
    for (int module_index = 0; module_index < modules->num; module_index++)
    {
      Module* module_cur = modules->arr[module_index];
      module_cur->setDepth(-1);
      module_cur->andTag(~TAG_OBJ_PROCESSED);
    }
    // calc the modules users
    for (int module_index = 0; module_index < modules->num; module_index++)
    {
      Module* module_cur = modules->arr[module_index];
      if (module_cur->instances)
      for (int instance_index = 0; instance_index < module_cur->instances->num; instance_index++)
      {
        Instance* instance_cur = module_cur->instances->arr[instance_index];
        Module* module_prototype = this->getModule(instance_cur->getModuleName());
        if (module_prototype)
          module_prototype->addUser(instance_cur);
      }
    }
    // calc the depth of the modules starting from the top level modules (which are not instantiated)
    for (int module_index = 0; module_index < modules->num; module_index++)
    {
      Module* module_cur = modules->arr[module_index];
      if (module_cur->users == NULL)
        this->findHierarchy(modules, module_cur, 0);
    }
    // calc the incidence of instances each module is built of
    for (int depth_cur = this->getMaxDepth(modules); depth_cur >= 0; depth_cur--)
    {
      for (int module_index = 0; module_index < modules->num; module_index++)
      {
        Module* module_cur = modules->arr[module_index];
        if (module_cur->testTag(TAG_MODULE_COMPLEX))
          if (module_cur->getDepth() == depth_cur)
            module_cur->calcIncidence();
      }
    }
  }
}


//---------------------------------------------------------------------------
int Lib::getMaxDepth(ModuleArray* modules)
{
  int depth_max = -1;
  if (modules)
  for (int module_index = 0; module_index < modules->num; module_index++)
  {
    Module* module_cur = modules->arr[module_index];
    int depth_cur = module_cur->getDepth();
    if (depth_cur > depth_max)
      depth_max = depth_cur;
  }
  return depth_max;
}

