%{
#include "messages.h"
#include "str_utils.h"
#include "sv_circuit.h"
#include "sverilog_y.h"

//#define LLDEBUG 1
#ifdef LLDEBUG
  #include <stdio.h>
  #define LLTRACE(msg) fprintf(stdout, msg)
#else
  #define LLTRACE(msg) // msg
#endif

char* sverilog_filename;
%}

%option yylineno

WS                 [\n\r\t\b ]
NWS                [^\n\r\t\b ]
COMMENT_C_OPEN     "/*"
COMMENT_C_CLOSE    "*/"
COMMENT_CPP        "//"
DIGIT              [0-9]
ALPHA              [a-zA-Z_]
ALPHANUM           [a-zA-Z_0-9]
INUMBER            {DIGIT}{DIGIT}*
RNUMBER1           {DIGIT}*\.{DIGIT}+
RNUMBER2           {DIGIT}+\.{DIGIT}*
RNUMBER            {RNUMBER1}|{RNUMBER2}
VDD                "1'b1"
GND                "1'b0"
ID_STANDART        {ALPHA}+({DIGIT}*{ALPHA}*)*
ID_ESCAPE          "\\"{NWS}+
ID                 {ID_STANDART}|{ID_ESCAPE}|{VDD}|{GND}
KEY_MODULE         module
KEY_ENDMODULE      endmodule
KEY_INPUT          input
KEY_OUTPUT         output
KEY_INOUT          inout
KEY_REG            reg
KEY_WIRE           wire
KEY_TRIREG         trireg
KEY_PARAMETER      parameter
KEY_ASSIGN         assign
KEY_IF             if
KEY_ELSE           else

KEY_SPECIFY        specify
KEY_ENDSPECIFY     endspecify

PRE_DEFINE         \`define
PRE_TIMESCALE      \`timescale
PRE_CELLDEFINE     \`celldefine
PRE_ENDCELLDEFINE  \`endcelldefine
PRE_DELAY_MODE_PATH \`delay_mode_path
PRE_IFDEF          \`ifdef
PRE_ENDIF          \`endif


%x COMMENT_C
%x IGNORE_LINE
%x IFDEF
%x SPECIFY

%%
%{
int comment_depth = 0;
int ifdef_depth = 0;
int specify_depth = 0;
%}

{WS}+              {
  /* eat white spaces */
}

{COMMENT_C_OPEN}   {
  LLTRACE("COMMENT_C_OPEN\n");
  comment_depth = 1;
  BEGIN COMMENT_C;
}

<COMMENT_C>{COMMENT_C_OPEN} {
  LLTRACE("COMMENT_C_OPEN\n");
  comment_depth++;
}

<COMMENT_C>{COMMENT_C_CLOSE} {
  LLTRACE("COMMENT_C_CLOSE\n");
  comment_depth--;
  if (comment_depth == 0)
    BEGIN INITIAL;
}

<COMMENT_C>.|\n    {
  /* eat comments */
}

{COMMENT_CPP}      {
  LLTRACE("COMMENT_CPP\n");
  BEGIN IGNORE_LINE;
}

<IGNORE_LINE>.     {
  /* eat line */
}

<IGNORE_LINE>\n    {
  LLTRACE("IGNORE_LINE\n");
  BEGIN INITIAL;
}

{KEY_IF}           {
  BEGIN IGNORE_LINE;
}

{KEY_ELSE}         {
  BEGIN IGNORE_LINE;
}

{KEY_ASSIGN}       {
  BEGIN IGNORE_LINE;
}

{PRE_TIMESCALE}    {
  BEGIN IGNORE_LINE;
}

{PRE_CELLDEFINE}   {
  BEGIN IGNORE_LINE;
}

{PRE_ENDCELLDEFINE} {
  BEGIN IGNORE_LINE;
}

{PRE_DEFINE}       {
  BEGIN IGNORE_LINE;
}

{PRE_DELAY_MODE_PATH} {
  BEGIN IGNORE_LINE;
}

{KEY_TRIREG}       {
  BEGIN IGNORE_LINE;
}

{PRE_IFDEF}        {
  LLTRACE("PRE_IFDEF\n");
  ifdef_depth = 1;
  BEGIN IFDEF;
}

<IFDEF>{PRE_IFDEF} {
  LLTRACE("PRE_IFDEF\n");
  ifdef_depth++;
  fprintf(stdout, "dpeth=%d", ifdef_depth);
}

<IFDEF>{PRE_ENDIF} {
  LLTRACE("PRE_ENDIF\n");
  ifdef_depth--;
  if (ifdef_depth == 0)
    BEGIN INITIAL;
}

<IFDEF>.|\n        {
  /* eat ifdef */
}



{KEY_SPECIFY}        {
  LLTRACE("KEY_SPECIFY\n");
  specify_depth = 1;
  BEGIN SPECIFY;
}

<SPECIFY>{KEY_SPECIFY} {
  LLTRACE("KEY_SPECIFY\n");
  specify_depth++;
}

<SPECIFY>{KEY_ENDSPECIFY} {
  LLTRACE("KEY_ENDSPECIFY\n");
  specify_depth--;
  if (specify_depth == 0)
    BEGIN INITIAL;
}

<SPECIFY>.|\n        {
  /* eat ifdef */
}


{KEY_MODULE}       {
  LLTRACE("KEY_MODULE\n");
  return YYMODULE;
}

{KEY_ENDMODULE}    {
  LLTRACE("KEY_ENDMODULE\n");
  return YYENDMODULE;
}

{KEY_INPUT}        {
  LLTRACE("KEY_INPUT\n");
  return YYINPUT;
}

{KEY_OUTPUT}       {
  LLTRACE("KEY_OUTPUT\n");
  return YYOUTPUT;
}

{KEY_INOUT}        {
  LLTRACE("KEY_INOUT\n");
  return YYINOUT;
}

{KEY_REG}          {
  LLTRACE("KEY_REG\n");
  return YYREG;
}

{KEY_WIRE}         {
  LLTRACE("KEY_WIRE\n");
  return YYWIRE;
}

{ID}               {
  LLTRACE("ID: "); LLTRACE(yytext); LLTRACE("\n");
  sverilog_lval.string = namecopy(yytext);
  return YYID;
}

{INUMBER}          {
  LLTRACE("INUMBER: "); LLTRACE(yytext); LLTRACE("\n");
  sverilog_lval.inumber = atoi(yytext);
  return YYINUMBER;
}

{RNUMBER}          {
  LLTRACE("RNUMBER: "); LLTRACE(yytext); LLTRACE("\n");
  sverilog_lval.rnumber = atof(yytext);
  return YYRNUMBER;
}

.                  {
  LLTRACE(yytext); LLTRACE("\n");
  return(yytext[0]);
}

<<EOF>>            {
  LLTRACE("<EOF>\n");
  yyterminate();
  return 0;
}

%%

int yywrap()
{
  sverilog_lineno = 0;
  // Free the buffer which remains un-freed due to a flex bug
  yy_delete_buffer(yy_current_buffer);
  yy_init = 1;
  return 1;
}

void sverilog_error(char* message)
{
  Error("%s in line %d in file <%s>", message, sverilog_lineno, sverilog_filename);
}

