#include <stdio.h>
#include "sv_module.h"
#include "sv_object.h"
#include "sv_parameter.h"
#include "sv_port.h"
#include "sv_instance.h"
#include "sv_lib.h"
#include "messages.h"

//---------------------------------------------------------------------------
Module::Module(int tag, char* name, ParameterArray* parameters, PortArray* ports, InstanceArray* instances, Lib* lib) : VerilogObject(tag, name)
{
  this->depth = -1;
  this->width = 0;

  this->users = NULL;

  this->parameters = NULL;
  this->addParameters(parameters);

  this->ports = NULL;
  this->addPorts(ports);

  this->port_vdd = new Port(TAG_PORT_WIRE | TAG_NET_VDD | TAG_OBJ_HIDDEN, namecopy(NAME_NET_VDD), NULL, NULL, NULL);
  this->addPort(this->port_vdd);

  this->port_gnd = new Port(TAG_PORT_WIRE | TAG_NET_GND | TAG_OBJ_HIDDEN, namecopy(NAME_NET_GND), NULL, NULL, NULL);
  this->addPort(this->port_gnd);

  this->instances = NULL;
  this->addInstances(instances);

  this->lib = lib;
}

//---------------------------------------------------------------------------
Module::Module(const Module &module) : VerilogObject(module)
{
  this->depth = module.depth;
  this->width = module.width;
  // copy users
  this->users = NULL;
  this->addUsers(module.users);
  // copy parameters
  this->parameters = NULL;
  if (module.parameters)
  {
    this->parameters = new ParameterArray(module.parameters->num);
    for (int parameter_index = 0; parameter_index < module.parameters->num; parameter_index++)
      this->addParameter(new Parameter(*module.parameters->arr[parameter_index]));
  }
  // copy ports
  this->ports = NULL;
  if (module.ports)
  {
    this->ports = new PortArray(module.ports->num);
    for (int port_index = 0; port_index < module.ports->num; port_index++)
      this->addPort(new Port(*module.ports->arr[port_index]));
  }
  // find vdd
  this->port_vdd = NULL;
  if (module.port_vdd)
    this->port_vdd = this->getPort(module.port_vdd->getName());
  // find gnd
  this->port_gnd = NULL;
  if (module.port_gnd)
    this->port_gnd = this->getPort(module.port_gnd->getName());
  // copy instances
  this->instances = NULL;
  if (module.instances)
  {
    this->instances = new InstanceArray(module.instances->num);
    for (int instance_index = 0; instance_index < module.instances->num; instance_index++)
      this->addInstance(new Instance(*module.instances->arr[instance_index]));
  }
  // copy lib
  this->lib = lib;
  // recalculate connection nets to this module ports
  this->calcConnections();
}

//---------------------------------------------------------------------------
Module::~Module()
{
  clean();
  delete this->users; this->users = NULL;
  delete this->instances; this->instances = NULL;
  delete this->parameters; this->parameters = NULL;
  delete this->ports; this->ports = NULL;
}

//---------------------------------------------------------------------------
void Module::clean()
{
  if (this->instances) this->instances->removeAll(true);
  if (this->parameters) this->parameters->removeAll(true);
  if (this->ports) this->ports->removeAll(true);
}

//---------------------------------------------------------------------------
void Module::setDepth(int depth)
{
  this->depth = depth;
}

//---------------------------------------------------------------------------
int Module::getDepth()
{
  return this->depth;
}

//---------------------------------------------------------------------------
void Module::setWidth(int width)
{
  this->width = width;
}

//---------------------------------------------------------------------------
int Module::getWidth()
{
  return this->width;
}

//---------------------------------------------------------------------------
void Module::addUser(Instance* user, int position)
{
  if (!this->users)
    this->users = new InstanceArray(1);
  if (position < 0)
    this->users->add(user);
  else
    this->users->addAt(position, user);
}

//---------------------------------------------------------------------------
void Module::addUsers(InstanceArray* users)
{
  if (users)
  for (int user_index = 0; user_index < users->num; user_index++)
  {
    Instance* user_cur = users->arr[user_index];
    this->addUser(user_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeUser(Instance* user)
{
  this->users->remove(user, false);
}

//---------------------------------------------------------------------------
void Module::addParameter(Parameter* parameter, int position)
{
  if (!this->parameters)
    this->parameters = new ParameterArray(1);
  if (position < 0)
    this->parameters->add(parameter);
  else
    this->parameters->addAt(position, parameter);
}

//---------------------------------------------------------------------------
void Module::addParameters(ParameterArray* parameters)
{
  if (parameters)
  for (int parameter_index = 0; parameter_index < parameters->num; parameter_index++)
  {
    Parameter* parameter_cur = parameters->arr[parameter_index];
    this->addParameter(parameter_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeParameter(Parameter* parameter)
{
  this->parameters->remove(parameter, true);
}

//---------------------------------------------------------------------------
Parameter* Module::getParameter(char* parameter_name)
{
  if (this->parameters)
  {
    Parameter* parameter_mask = new Parameter(TAG_ALL, namecopy(parameter_name));
    int parameter_index = this->parameters->find(parameter_mask, 0, VerilogObject::compareByMask);
    delete parameter_mask;
    if (parameter_index >= 0)
      return  this->parameters->arr[parameter_index];
  }
  return NULL;
}

//---------------------------------------------------------------------------
void Module::addPort(Port* port, int position)
{
  if (!this->ports)
    this->ports = new PortArray(1);
  if (position < 0)
    this->ports->add(port);
  else
    this->ports->addAt(position, port);
}

//---------------------------------------------------------------------------
void Module::addPorts(PortArray* ports)
{
  if (ports)
  for (int port_index = 0; port_index < ports->num; port_index++)
  {
    Port* port_cur = ports->arr[port_index];
    this->addPort(port_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removePort(Port* port)
{
  this->ports->remove(port, true);
}

//---------------------------------------------------------------------------
Port* Module::getPort(char* port_name)
{
  if (this->ports)
  {
    Port* port_mask = new Port(TAG_ALL, namecopy(port_name));
    int port_index = this->ports->find(port_mask, 0, VerilogObject::compareByMask);
    delete port_mask;
    if (port_index >= 0)
      return  this->ports->arr[port_index];
  }
  return NULL;
}

//---------------------------------------------------------------------------
int Module::getPortNum(int tag_mask)
{
  int port_num = 0;
  if (this->ports)
  for (int port_index = 0; port_index < this->ports->num; port_index++)
  {
     Port* port_cur = this->ports->arr[port_index];
     if (!port_cur->testTag(TAG_OBJ_HIDDEN))
       if (port_cur->testTag(tag_mask))
         port_num++;
  }
  return port_num;
}

//---------------------------------------------------------------------------
Net* Module::getNet(char* net_name, int net_index)
{
  Port* port_cur = getPort(net_name);
  if (port_cur)
    return port_cur->getNet(net_name, net_index);
  else
    return NULL;
}

//---------------------------------------------------------------------------
Net* Module::getNet(Net* net)
{
  return this->getNet(net->getName(), net->getIndex());
}

//---------------------------------------------------------------------------
void Module::addInstance(Instance* instance, int position)
{
  if (!this->instances)
    this->instances = new InstanceArray(1);
  instance->setContainer(this);
  if (position < 0)
    this->instances->add(instance);
  else
    this->instances->addAt(position, instance);
}

//---------------------------------------------------------------------------
void Module::addInstances(InstanceArray* instances)
{
  if (instances)
  for (int instance_index = 0; instance_index < instances->num; instance_index++)
  {
    Instance* instance_cur = instances->arr[instance_index];
    this->addInstance(instance_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeInstance(Instance* instance)
{
  this->instances->remove(instance, true);
}

//---------------------------------------------------------------------------
Instance* Module::getInstance(char* instance_name, int instance_index)
{
  if (this->instances)
  {
    Instance* instance_mask = new Instance(TAG_ALL, namecopy(instance_name));
    int instance_find_index = this->instances->find(instance_mask, 0, VerilogObject::compareByMask);
    delete instance_mask;
    if (instance_find_index >= 0)
    {
      Instance* instance = this->instances->arr[instance_find_index];
      if (instance_index == instance->getIndex())
        return instance;
    }
  }
  return NULL;
}


//---------------------------------------------------------------------------
void Module::calcConnections()
{
//Message("module <%s>", this->getName());
  NetArray* nets_tmp = new NetArray(1);
  if (this->instances)
  for (int instance_index = 0; instance_index < this->instances->num; instance_index++)
  {
    Instance* instance_cur = this->instances->arr[instance_index];
    char* instance_full_name = instance_cur->genFullName();
//Message("  instance <%s> of module <%s>", instance_full_name, instance_cur->getModuleName());
    if (instance_cur->connections)
    for (int connection_index = 0; connection_index < instance_cur->connections->num; connection_index++)
    {
      Connection* connection_cur = instance_cur->connections->arr[connection_index];
//Message("    connection <%s>", connection_cur->getName());
      if (connection_cur->nets)
      for (int net_index = 0; net_index < connection_cur->nets->num; net_index++)
      {
        Net* net_cur = connection_cur->nets->arr[net_index];
        if (net_cur)
        {
//char* net_full_name = net_cur->genLabelName();
//Message("      net <%s>", net_full_name);
//delete [] net_full_name;
          bool net_undefined = false;
          Net* net_new = NULL;
          Port* port_cur = this->getPort(net_cur->getName());
          if (!port_cur)
            if (net_cur->getIndex() == -1)
            {
              Warning("in module <%s> instance <%s> is connected to undeclared net <%s>", 
                this->name, instance_full_name, net_cur->getName());
              port_cur = new Port(TAG_PORT_WIRE, namecopy(net_cur->getName()), NULL, NULL, NULL);
              this->addPort(port_cur);
            }
          if (port_cur)
          {
            // connection of a bus
            if (net_cur->getIndex() == -1)
            {
              for (int net_new_index = 0; net_new_index < port_cur->nets->num; net_new_index++)
              {
                net_new = port_cur->nets->arr[net_new_index];
                if (net_new != net_cur)
                {
                  if (net_new_index == 0)
                    connection_cur->nets->arr[net_index] = net_new;
                  else
                    connection_cur->nets->addAt(++net_index, net_new);
                }
              }
            }
            // connection of a wire
            else
            {
              net_new = port_cur->getNet(net_cur->getName(), net_cur->getIndex());
              if (net_new != net_cur)
              {
                if (net_new)
                  connection_cur->nets->arr[net_index] = net_new;
                else
                  net_undefined = true;
              }
            }
          }
          else
            net_undefined = true;
          // remember the nets which can be removed in the end  
          if (net_cur != net_new)
          {
//            Message("module <%s> instance <%s> connection <%s> net <%s>", this->name, instance_cur->getName(), connection_cur->getName(), net_cur->getName());
            if ( nets_tmp->locate(net_cur) < 0 )
               nets_tmp->add(net_cur);
          }
 
          if (net_undefined)
          {
            char* net_full_name = net_cur->genFullName();
            Error("in module <%s> instance <%s> is connected to undeclared net <%s> which is in a bus",
              this->name, instance_full_name, net_full_name);
            delete [] net_full_name;
          }
        }
      }
    }
    delete [] instance_full_name;
  }
  if (nets_tmp)
  {
    nets_tmp->removeAll(true);
    delete nets_tmp; nets_tmp = NULL;
  }
}

//---------------------------------------------------------------------------
void Module::sortConnections()
{
  if (this->instances)
  for (int instance_index = 0; instance_index < this->instances->num; instance_index++)
  {
    Instance* instance_cur = this->instances->arr[instance_index];
    char* instance_full_name = instance_cur->genFullName();
    if (instance_full_name)
    {
      if (instance_cur->connections)
      for (int connection_index = 0; connection_index < instance_cur->connections->num; connection_index++)
      {
        Connection* connection_cur = instance_cur->connections->arr[connection_index];
        if (connection_cur->nets)
        for (int net_index = 0; net_index < connection_cur->nets->num; net_index++)
        {
          Net* net_cur = connection_cur->nets->arr[net_index];
          if (net_cur)
          {
            Port* port_prot = this->getPort(net_cur->getName());
            if (port_prot)
              net_cur->setTag(port_prot->getTag());
          }
        }
      }
      //
      Module* module_prot = this->lib->getModule(instance_cur->getModuleName());
      if (module_prot)
      // if prototype is found
      {
        if (instance_cur->connections)
        {
          for (int connection_index = 0; connection_index < instance_cur->connections->num; connection_index++)
          {
            Connection* connection_cur = instance_cur->connections->arr[connection_index];
            if (connection_cur->getName() == NULL)
            {
              if (module_prot->ports)
              {
                int port_index = -1;
                int port_connection_index = port_index;
                while (port_connection_index < connection_index)
                {
                  port_index++;
                  if (port_index < module_prot->ports->num)
                  {
                    Port* port_cur = module_prot->ports->arr[port_index];
                    if (!port_cur->testTag(TAG_PORT_WIRE) &&
                        !port_cur->testTag(TAG_PORT_REG) &&
                        !port_cur->testTag(TAG_OBJ_HIDDEN))
                      port_connection_index++;
                  }
                  else
                    Error("instance <%s> of module <%s> in module <%s> has more connections then ports available",
                      instance_full_name, instance_cur->getModuleName(), this->name);
                }
                if (port_connection_index == connection_index)
                {
                  Port* port_cur = module_prot->ports->arr[port_index];
                  connection_cur->setTag(port_cur->getTag());
                  connection_cur->setName(namecopy(port_cur->getName()));
                }
              }
            }
          }
          // sort connections according to port order
          if (module_prot->ports)
          {
            int connection_index_max = 0;
            for (int port_index = 0; port_index < module_prot->ports->num; port_index++)
            {
              Port* port_cur = module_prot->ports->arr[port_index];
              if (!port_cur->testTag(TAG_PORT_WIRE) &&
                  !port_cur->testTag(TAG_PORT_REG) &&
                  !port_cur->testTag(TAG_OBJ_HIDDEN))
              {
                bool connection_found = false;
                for (int connection_index = connection_index_max; (connection_index < instance_cur->connections->num) && (!connection_found); connection_index++)
                {
                  Connection* connection_cur = instance_cur->connections->arr[connection_index];
                  if (connection_cur->getName())
                  if (strcmp(port_cur->getName(), connection_cur->getName()) == 0)
                  {
                    connection_found = true;
                    instance_cur->connections->arr[connection_index] = instance_cur->connections->arr[connection_index_max];
                    instance_cur->connections->arr[connection_index_max] = connection_cur;
                    connection_cur->setTag(port_cur->getTag());
                    connection_index_max++;
                  }
                }
                if (!connection_found)
                {
                  Connection* connection_new = new Connection(port_cur->getTag(), namecopy(port_cur->getName()));
                  instance_cur->addConnection(connection_new, connection_index_max);
                  connection_index_max++;
                }
              }
            }
          }
        }
      }
      else
        Error("instance <%s> of module <%s> in module <%s> has no prototype!!!", 
          instance_full_name, instance_cur->getModuleName(), this->name);
      
      for (int connection_index = 0; connection_index < instance_cur->connections->num; connection_index++)
      {
        Connection* connection_cur = instance_cur->connections->arr[connection_index];
        if (connection_cur->getName() == NULL)
        {
          Error("some ports of instance <%s> of module <%s> in module <%s> cannot be associated", 
            instance_full_name, instance_cur->getModuleName(), this->name);
          if (connection_index == (instance_cur->connections->num - 1))
            connection_cur->setTag(TAG_PORT_OUTPUT);
          else
            connection_cur->setTag(TAG_PORT_INPUT);
        }
      }
    }
    delete [] instance_full_name;
  }
}

//---------------------------------------------------------------------------
void Module::calcIncidence()
{
  // for every module port find its connections
  if (this->ports)
  for (int port_index = 0; port_index < this->ports->num; port_index++)
  {
    Port* port_cur = this->ports->arr[port_index];
    if (port_cur->nets)
    for (int net_index = 0; net_index < port_cur->nets->num; net_index++)
    {
      Net* net_cur = port_cur->nets->arr[net_index];
      delete net_cur->preset; net_cur->preset = NULL;
      delete net_cur->postset; net_cur->postset = NULL;
      if (this->instances)
      for (int instance_index = 0; instance_index < this->instances->num; instance_index++)
      {
        Instance* instance_cur = this->instances->arr[instance_index];
        if (instance_cur->connections)
        for (int connection_index = 0; connection_index < instance_cur->connections->num; connection_index++)
        {
          Connection* connection_cur = instance_cur->connections->arr[connection_index];
          if (connection_cur->getNet(net_cur->getName(), net_cur->getIndex()))
          {
            if (connection_cur->testTag(TAG_PORT_INPUT))
            {
              net_cur->addPostset(instance_cur);
            }
            if (connection_cur->testTag(TAG_PORT_OUTPUT))
            {
              net_cur->addPreset(instance_cur);
            }
          }
        }
      }
    }
  }
}
